package org.ids_mannheim;

import org.anarres.parallelgzip.ParallelGZIPOutputStream;
import picocli.CommandLine;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;


@CommandLine.Command(mixinStandardHelpOptions = true,
        name = "totalngram", description = "add ngram counts from KorAP-XML or CoNLL-U files")

public class TotalNGram implements Callable<Integer> {
    static public final Logger logger;
    private static final int MAX_THREADS = Runtime.getRuntime().availableProcessors() * 2 / 3;
    private static final String DEFAULT_LOGFILE = "totalngram.log";

    static {
        String path = TotalNGram.class.getClassLoader()
                .getResource("logging.properties")
                .getFile();
        System.setProperty("java.util.logging.config.file", path);
        logger = Logger.getLogger(TotalNGram.class.getSimpleName());
    }

    @CommandLine.Parameters(arity = "1..*", description = "input files")
    private ArrayList<String> inputFiles = new ArrayList<>();
    @CommandLine.Option(names = {"-L",
            "--log-file"}, defaultValue = DEFAULT_LOGFILE, description = "log file name (default: ${DEFAULT-VALUE})")
    String logFileName;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-o",
            "--output-file"}, description = "Output file (default: ${DEFAULT-VALUE})")
    String output_fillename = "-";

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"--force"}, description = "Force overwrite (default: ${DEFAULT-VALUE})")
    boolean force_overwrite = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-l", "--with-lemma-pos"}, description = "Use also lemma and part-of-speech annotations (default: ${DEFAULT-VALUE}")
    boolean with_lemma_and_pos = false;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-P",
            "--max-procs"}, description = "Run up to max-procs processes at a time (default: ${DEFAULT-VALUE})")
    int max_threads = MAX_THREADS;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-p",
            "--worker-pool"}, description = "Run preprocessing on extern hosts, e.g. '10*local,5*host1,3*smith@host2' (default: ${DEFAULT-VALUE})")
    String worker_pool_specification = "";

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-n",
            "--ngram-size"}, description = "n-gram size (default: ${DEFAULT-VALUE})")
    int ngram_size = 1;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-F",
            "--folds"}, description = "number of folds (default: ${DEFAULT-VALUE})")
    int FOLDS = 10;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-f",
            "--fold"}, description = "current folds (default: ${DEFAULT-VALUE})")
    int fold = 1;

    private Progressbar etaPrinter;

    public TotalNGram() {
    }

    public static void main(String[] args) {
        System.exit(new CommandLine(new TotalNGram()).execute(args));
    }

    private void initLogging() {
        if (logFileName.equals(DEFAULT_LOGFILE) && !output_fillename.equals("-")) {
            logFileName = Utils.newExtension(output_fillename, "log");
        }
        try {
            FileHandler fileHandler = new FileHandler(logFileName);
            System.setProperty("java.util.logging.SimpleFormatter.format",
                    "[%1$tF %1$tT] [%4$-7s] %5$s %n");
            fileHandler.setFormatter(new SimpleFormatter());
            //java.util.logging.SimpleFormatter.format="%1$tb %1$td, %1$tY %1$tl:%1$tM:%1$tS %1$Tp %2$s%n%4$s: %5$s%n";
            java.util.Locale.setDefault(Locale.Category.FORMAT, Locale.ROOT);
            logger.addHandler(fileHandler);
            ConsoleHandler consoleHandler = new ConsoleHandler();
            consoleHandler.setLevel(Level.WARNING);
            logger.addHandler(consoleHandler);
        } catch (Exception e) {
            System.err.println("Cannot write logfile " + logFileName + ": " + e);
            System.exit(-1);
        }
    }

    @Override
    public Integer call() throws Exception {
        initLogging();
        final PrintStream output_stream;
        if ((output_fillename == null) || output_fillename.equals("-")) {
            output_stream = System.out;
        } else {
            File f = Utils.createFile(output_fillename, force_overwrite);
            if (output_fillename.endsWith(".gz")) {
                output_stream = new PrintStream(new ParallelGZIPOutputStream((new BufferedOutputStream(new FileOutputStream(f)))));
            } else {
                output_stream = new PrintStream(new BufferedOutputStream(new FileOutputStream(f)));
            }

        }

        ConcurrentHashMap<String, AtomicInteger> map = new ConcurrentHashMap<>();

        long totalFilesSizes = inputFiles.parallelStream().mapToLong(fname -> {
            long l = new File(fname).length();
            if (l == 0) {
                logger.severe("File "+fname+" does not exist or is empty.");
                System.exit(-1);
            }
            return l;
        }).sum();
        inputFiles = (ArrayList<String>) inputFiles.parallelStream().sorted((a, b) ->
           Long.compareUnsigned(new File(b).length(), new File(a).length())
        ).collect(Collectors.toList());

        etaPrinter = new Progressbar(totalFilesSizes);
        BlockingQueue<Integer> queue = new LinkedBlockingQueue<>(inputFiles.size());
        ExecutorService es = Executors.newCachedThreadPool();
        WorkerNodePool workerNodePool = new WorkerNodePool(worker_pool_specification);
        if (!worker_pool_specification.equals("")) {
            max_threads = workerNodePool.size;
        }
        int threads = Math.min(max_threads, inputFiles.size());
        logger.info("Processing fold " + fold + "/" + FOLDS);
        logger.info("Using " + threads + " threads");
        IntStream.range(0, threads).forEach(unused -> es.execute(new Worker(queue, inputFiles, ngram_size, fold, FOLDS, map, with_lemma_and_pos, workerNodePool, etaPrinter, logger)));
        queue.addAll(IntStream.range(0, inputFiles.size()).boxed().collect(Collectors.toList()));
        IntStream.range(0, threads).forEach(unused -> {
            try {
                queue.put(-1);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        });
        es.shutdown();
        boolean finished = es.awaitTermination(120, TimeUnit.HOURS);
        etaPrinter.finish();
        logger.info("Sorting and writing frequency table.");
        System.err.println("Sorting and writing frequency table.");
        map.entrySet().parallelStream()
                .sorted((a, b) -> {
                    int cmp1 = Integer.compareUnsigned(b.getValue().get(), a.getValue().get());
                    if (cmp1 != 0) {
                        return cmp1;
                    } else {
                        return a.getKey().compareTo(b.getKey());
                    }
                })
                .forEachOrdered(entry -> output_stream.println(entry.getKey() + "\t" + entry.getValue().toString()));
        logger.info("Finished.");
        output_stream.close();
        return null;
    }
}
